//
// VNC Viewer for B-right/V
// (C)Copyright 2003, Yashiro Takeshi
//

// ----------------------------------------------------
//  RFB ץȥط
// ----------------------------------------------------

#define RFBTRANSACTIONS_IMPL
#include "rfbTransactions.h"

#include <bstdio.h>
#include <tstring.h>
#include <keycode.h>
#include <unix/zlib.h>
#include "ypoint.h"
#include "ygraphic.h"
#include "ystring.h"
#include "ylockobject.h"
#include "errorpanel.h"
#include "vncauth.h"
#include "VncWindow.h"

static char* keyTable = 0;

// ----------------------------------------------------
//  VNC: С
// ----------------------------------------------------
bool execVNCProtocolVersion(YLibDraft::YTCPSocket* sock, int* ver, int* subver){
  char msg[13];
  // receive from server
  sock->read(msg, 12);
  if (msg[0] == 'R' && msg[1] == 'F' && msg[2] == 'B' && msg[3] == ' '
      && msg[11] == '\x0a' && msg[7] == '.'){
    if (ver != 0){
      *ver = (msg[4]-'0')*100 + (msg[5]-'0')*10 + (msg[6]-'0');
    }
    if (subver != 0){
      *subver = (msg[8]-'0')*100 + (msg[9]-'0')*10 + (msg[10]-'0');
    }
    // reply to server
    msg[4] = msg[5] = msg[8] = msg[9] = '0';
    msg[6] = msg[10] = '3';
    sock->write(msg, 12);
    return true;
  }
  return false;
}

// ----------------------------------------------------
//  VNC: ǧڼ¹
// ----------------------------------------------------
int execVNCAuthentication(YLibDraft::YTCPSocket* sock, char* passwd)
{
  unsigned char challenge[17];
  sock->read((B*)challenge, 16);
  vncEncryptBytes(challenge, passwd);
  sock->write((B*)challenge, 16);
  unsigned long status;
  sock->read((B*)&status, 4);
  return htonl(status);
}

// ----------------------------------------------------
//  VNC: н
// ----------------------------------------------------
TC* execVNCServerInitalization(YLibDraft::YTCPSocket* sock, rfbServerInitMsg *msg)
{
  sock->read((B*)msg, sizeof(rfbServerInitMsg));
  msg->framebufferWidth = ntohs(msg->framebufferWidth);
  msg->framebufferHeight = ntohs(msg->framebufferHeight);
  msg->format.redMax = ntohs(msg->format.redMax);
  msg->format.greenMax = ntohs(msg->format.greenMax);
  msg->format.blueMax = ntohs(msg->format.blueMax);
  msg->nameLength = ntohl(msg->nameLength);
  TC* name = 0;
  if (msg->nameLength > 0){
    char* buf = new char[msg->nameLength+1];
    name = new TC[(msg->nameLength+1)*sizeof(TC)];
    buf[msg->nameLength] = '\0';
    sock->read(buf, msg->nameLength);
    eucstotcs(name, buf);
    delete buf;
  }
  return name; // newƤΤ
}

// ----------------------------------------------------
//  VNC: 󥳡
// ----------------------------------------------------
void execVNCSetEncoding(YLibDraft::YTCPSocket* sock)
{
  const static char allowedTypes[] = { rfbEncodingHextile, rfbEncodingCopyRect, rfbEncodingRRE, rfbEncodingCoRRE, rfbEncodingRaw };
  rfbSetEncodingsMsg set_enc;
  set_enc.type = rfbSetEncodings;
  set_enc.nEncodings = htons(sizeof(allowedTypes) / sizeof(char));
  sock->write((B*)&set_enc, sizeof(rfbSetEncodingsMsg));

  for (int i = 0; i < sizeof(allowedTypes) / sizeof(char); i++){
    unsigned long v = htonl(allowedTypes[i]);
    sock->write((B*)&v, 4);
  }
}

// ----------------------------------------------------
//  VNC: ԥ
// ----------------------------------------------------
void execVNCSetPixelFormat(YLibDraft::YTCPSocket* sock, rfbServerInitMsg* server)
{
  rfbSetPixelFormatMsg msg;
  server->format.trueColour = 1;
  msg.type = rfbSetPixelFormat;
  msg.format = server->format;
  msg.format.redMax = htons(msg.format.redMax);
  msg.format.greenMax = htons(msg.format.greenMax);
  msg.format.blueMax = htons(msg.format.blueMax);
  sock->write((B*)&msg, sizeof(rfbSetPixelFormatMsg));
}

// ----------------------------------------------------
//  VNC: ޥ٥Ȥ
// ----------------------------------------------------
void execVNCMouseEvent(YLibDraft::YTCPSocket* sock, YSemaphore* sem, const YPoint& pt, char sts)
{
  rfbPointerEventMsg msg;
  msg.type = rfbPointerEvent;
  msg.buttonMask = sts;
  msg.x = htons(pt.x);
  msg.y = htons(pt.y);
  {
    YLockObject guard(sem);
    sock->write((B*)&msg, sizeof(rfbPointerEventMsg));
  }
}

// ----------------------------------------------------
//  VNC: ̹
// ----------------------------------------------------
void requestVNCFramebufferUpdate(YLibDraft::YTCPSocket* sock, YSemaphore* sem, const YRect &rect, bool first = false)
{
  rfbFramebufferUpdateRequestMsg msg;
  msg.type = rfbFramebufferUpdateRequest;
  msg.incremental = (first ? 0 : 1);
  msg.x = htons(rect.left);
  msg.y = htons(rect.top);
  msg.w = htons(rect.width());
  msg.h = htons(rect.height());
  {
    YLockObject guard(sem);
    sock->write((B*)&msg, sizeof(rfbFramebufferUpdateRequestMsg));
  }
}

// ----------------------------------------------------
//  VNC: ԥͤCOLORؤѴ
// ----------------------------------------------------
// not 'extern'ed as of yet
COLOR convVNCPixeltoCOLOR(char* buf, int bpp, const rfbServerInitMsg* server)
{
  unsigned int r, g, b;
  
  switch (bpp) {
  case 1:
    {
      unsigned char val = *((unsigned char *)buf);
      r = (val >> server->format.redShift) & server->format.redMax;
      g = (val >> server->format.greenShift) & server->format.greenMax;
      b = (val >> server->format.blueShift) & server->format.blueMax;
    }
    break;
  case 2:
    {
      unsigned short val = *((unsigned short *)buf);
      if (server->format.bigEndian)
        val = ntohs(val);
      r = (val >> server->format.redShift) & server->format.redMax;
      g = (val >> server->format.greenShift) & server->format.greenMax;
      b = (val >> server->format.blueShift) & server->format.blueMax;
    }
    break;
  case 4:
    {
      unsigned long val = *((unsigned long *)buf);
      if (server->format.bigEndian)
        val = ntohl(val);
      r = (val >> server->format.redShift) & server->format.redMax;
      g = (val >> server->format.greenShift) & server->format.greenMax;
      b = (val >> server->format.blueShift) & server->format.blueMax;
    }
    break;
  default:
    // error
    r = g = b = 0;
  };
  return 0x10000000
    | (((r * 255) / server->format.redMax) << 16)
    | (((g * 255) / server->format.greenMax) << 8)
    | ((b * 255) / server->format.blueMax);
}

// ----------------------------------------------------
//  VNC: RawEncodingˤե졼Хåեι
// ----------------------------------------------------
// not 'extern'ed as of yet
void execRawEncodingUpdate(YLibDraft::YTCPSocket *sock, rfbServerInitMsg *server, W gid, const YRect& rect)
{
  int bpp = server->format.bitsPerPixel / 8;
  int bsize = rect.width() * bpp;
  char* buf = new char[bsize+1];
  
  YPoint pnt;
  
  for (pnt.y = rect.top; pnt.y < rect.bottom; pnt.y++){
    sock->read((B*)buf, bsize);
    int pos = 0;
    for (pnt.x = rect.left; pnt.x < rect.right; pnt.x++){
      COLOR col = convVNCPixeltoCOLOR((B*)(buf+pos), bpp, server);
      gdra_pnt(gid, pnt, col, G_STORE);
      pos += bpp;
    }
  }
  delete buf;
}

// ----------------------------------------------------
//  VNC: CopyRectEncodingˤե졼Хåեι
// ----------------------------------------------------
// not 'extern'ed as of yet
void execCopyRectEncodingEncodingUpdate(YLibDraft::YTCPSocket *sock, W gid, const YRect& rect)
{
  YRect src;
  rfbCopyRect msg;
  sock->read((B*)&msg, sizeof(rfbCopyRect));
  src.left = ntohs(msg.srcX);
  src.top = ntohs(msg.srcY);
  src.right = src.left + rect.width();
  src.bottom = src.top + rect.height();
  gcop_bmp(gid, (RECT*)&src, gid, (RECT*)&rect, NULL, G_STORE);
}

// ----------------------------------------------------
//  VNC: RREEncodingˤե졼Хåեι
// ----------------------------------------------------
// not 'extern'ed as of yet
void execRREEncodingUpdate(YLibDraft::YTCPSocket *sock, YSemaphore* disp_sem, rfbServerInitMsg *server, W gid, const YRect& rect)
{
  PAT pat;
  unsigned long len;
  char buf[5];
  sock->read((B*)&len, 4);
  len = ntohl(len);
  int bpp = server->format.bitsPerPixel / 8;
  sock->read((B*)buf, bpp);
  pat.spat.kind = 0;
  pat.spat.hsize = pat.spat.vsize = 1;
  pat.spat.fgcol = convVNCPixeltoCOLOR((B*)buf, bpp, server);
  pat.spat.bgcol = 0x10000000;
  pat.spat.mask = FILL100;

  {
    YLockObject guard(disp_sem);
    gfil_rec(gid, rect, &pat, 0, G_STORE);
    
    for (int i = 0; i < len; i++){
      YRect subrect;
      rfbRectangle msg_rect;
      sock->read((B*)buf, bpp);
      pat.spat.fgcol = convVNCPixeltoCOLOR((B*)buf, bpp, server);
      sock->read((B*)&msg_rect, sizeof(rfbRectangle));
      subrect.left = rect.left + ntohs(msg_rect.x);
      subrect.top = rect.top + ntohs(msg_rect.y);
      subrect.right = subrect.left + ntohs(msg_rect.w);
      subrect.bottom = subrect.top + ntohs(msg_rect.h);
      gfil_rec(gid, subrect, &pat, 0, G_STORE);
    }
  }
}

// ----------------------------------------------------
//  VNC: CoRREEncodingˤե졼Хåեι
// ----------------------------------------------------
// not 'extern'ed as of yet
void execCoRREEncodingUpdate(YLibDraft::YTCPSocket *sock, YSemaphore* disp_sem, rfbServerInitMsg *server, W gid, const YRect& rect)
{
  PAT pat;
  unsigned long len;
  char buf[5];
  sock->read((B*)&len, 4);
  len = ntohl(len);
  int bpp = server->format.bitsPerPixel / 8;
  sock->read((B*)buf, bpp);
  pat.spat.kind = 0;
  pat.spat.hsize = pat.spat.vsize = 1;
  pat.spat.fgcol = convVNCPixeltoCOLOR((B*)buf, bpp, server);
  pat.spat.bgcol = 0x10000000;
  pat.spat.mask = FILL100;

  {  
    YLockObject guard(disp_sem);
    gfil_rec(gid, rect, &pat, 0, G_STORE);
    
    for (int i = 0; i < len; i++){
      YRect subrect;
      rfbCoRRERectangle msg_rect;
      sock->read((B*)buf, bpp);
      pat.spat.fgcol = convVNCPixeltoCOLOR((B*)buf, bpp, server);
      sock->read((B*)&msg_rect, sizeof(rfbCoRRERectangle));
      subrect.left = rect.left + msg_rect.x;
      subrect.top = rect.top + msg_rect.y;
      subrect.right = subrect.left + msg_rect.w;
      subrect.bottom = subrect.top + msg_rect.h;
      gfil_rec(gid, subrect, &pat, 0, G_STORE);
    }
  }
}

// ----------------------------------------------------
//  VNC: HextileEncodingˤե졼Хåեι
// ----------------------------------------------------
// not 'extern'ed as of yet
void execHextileEncodingUpdate(YLibDraft::YTCPSocket *sock, YSemaphore* disp_sem, rfbServerInitMsg *server, W gid, const YRect& rect)
{
  PAT pat;
  pat.spat.kind = 0;
  pat.spat.hsize = pat.spat.vsize = 1;
  pat.spat.mask = FILL100;
  int bpp = server->format.bitsPerPixel / 8;
  static W bgc = 0x90000000, fgc = 0x90000000;
  
  for (int y = rect.top; y < rect.bottom; y += 16){
    for (int x = rect.left; x < rect.right; x += 16){
      YRect subrect(x, y, x+16, y+16);
      subrect.right = ((subrect.right < rect.right) ?
                       subrect.right : rect.right);
      subrect.bottom = ((subrect.bottom < rect.bottom) ?
                        subrect.bottom : rect.bottom);
      unsigned char type, nrect;
      sock->read((B*)&type, 1);
      if (type == 0){
        pat.spat.fgcol = bgc;
        gfil_rec(gid, subrect, &pat, 0, G_STORE);
        continue;
      }
      if ((type & 1) != 0){
        // raw bit
        int bsize = subrect.width() * subrect.height() * bpp;
        char buf[1025];
        YPoint pnt;
        sock->read((B*)buf, bsize);
        int pos = 0;
        for (pnt.y = subrect.top; pnt.y < subrect.bottom; pnt.y++){
          for (pnt.x = subrect.left; pnt.x < subrect.right; pnt.x++){
            COLOR col = convVNCPixeltoCOLOR((B*)(buf+pos), bpp, server);
            gdra_pnt(gid, pnt, col, G_STORE);
            pos += bpp;
          }
        }
      }
      else {
        unsigned char buf[7];
        if ((type & 2) != 0){
          // background-specified bit
          sock->read((B*)buf, bpp);
          bgc = convVNCPixeltoCOLOR((B*)buf, bpp, server);
        }
        if ((type & 4) != 0){
          // foreground-specified bit
          sock->read((B*)buf, bpp);
          fgc = convVNCPixeltoCOLOR((B*)buf, bpp, server);
        }
        {
          YLockObject guard(disp_sem);
          
          pat.spat.fgcol = bgc;
          gfil_rec(gid, subrect, &pat, 0, G_STORE);
          if ((type & 8) != 0){
            // any-subrects bit
            YRect frac;
            sock->read((B*)&nrect, 1);
            if ((type & 16) != 0){
              // subrects-colored bit
              for (int i = 0; i < nrect; i++){
                sock->read((B*)buf, 2+bpp);
                pat.spat.fgcol = convVNCPixeltoCOLOR((B*)buf, bpp, server);
                frac.left = x + (buf[0+bpp] >> 4);
                frac.top = y + (buf[0+bpp] & 0x0fU);
                frac.right = frac.left + (buf[1+bpp] >> 4) + 1;
                frac.bottom = frac.top + (buf[1+bpp] & 0x0fU) + 1;
                gfil_rec(gid, frac, &pat, 0, G_STORE);
              }
            }
            else {
              pat.spat.fgcol = fgc;
              for (int i = 0; i < nrect; i++){ 
                sock->read((B*)buf, 2);
                frac.left = x + (buf[0] >> 4);
                frac.top = y + (buf[0] & 0x0fU);
                frac.right = frac.left + (buf[1] >> 4) + 1;
                frac.bottom = frac.top + (buf[1] & 0x0fU) + 1;
                gfil_rec(gid, frac, &pat, 0, G_STORE);
              }
            }
          }
        }
      }
    }
  }
}

#if 0
void execZRLEEncodingUpdate(YLibDraft::YTCPSocket *sock, YSemaphore* disp_sem, rfbServerInitMsg *server, W gid, const YRect& rect)
{
  z_stream z;
  z.zalloc = z.zfree = z.opaque = Z_NULL;

  z.next_in = Z_NULL;
  z.avail_in = 0;

  if (inflateInit(&z) != Z_OK){
    printf("zlib error: inflateInit\n");
    exit(-1);
  }

  sock->read(integer);
  sock->read(char * len);

#error "Unimplemented"
}
#endif

// ----------------------------------------------------
//  VNC: ե졼Хåեι
// ----------------------------------------------------
void execFramebufferUpdate(YLibDraft::YTCPSocket *sock, YSemaphore* disp_sem, rfbServerInitMsg *server, W wid, W gid)
{
  rfbFramebufferUpdateMsg msg;
  rfbFramebufferUpdateRectHeader rect_header;
  YRect rect;
  
  msg.type = rfbFramebufferUpdate; // not needed?
  sock->read((B*)(&msg.pad),
             sizeof(rfbFramebufferUpdateMsg) - sizeof(char));
  msg.nRects = ntohs(msg.nRects);
  
  for (int i = 0; i < msg.nRects; i++){
    sock->read((B*)&rect_header, sizeof(rfbFramebufferUpdateRectHeader));
    rect_header.encoding = ntohl(rect_header.encoding);
    rect.left = ntohs(rect_header.r.x);
    rect.top = ntohs(rect_header.r.y);
    rect.right = rect.left + ntohs(rect_header.r.w);
    rect.bottom = rect.top + ntohs(rect_header.r.h);
    
    switch (rect_header.encoding){
    case rfbEncodingRaw:
      //printf("rfbEncodingRaw\n");
      execRawEncodingUpdate(sock, server, gid, rect);
      break;
    case rfbEncodingCopyRect:
      //printf("rfbEncodingCopyRect\n");
      execCopyRectEncodingEncodingUpdate(sock, gid, rect);
      break;
    case rfbEncodingRRE:
      //printf("rfbEncodingRRE\n");
      execRREEncodingUpdate(sock, disp_sem, server, gid, rect);
      break;
    case rfbEncodingCoRRE:
      //printf("rfbCoEncodingRRE\n");
      execCoRREEncodingUpdate(sock, disp_sem, server, gid, rect);
      break;
    case rfbEncodingHextile:
      //printf("rfbHextileEncoding\n");
      execHextileEncodingUpdate(sock, disp_sem, server, gid, rect);
      break;
    default:
      {
        YString str((TC[]){L"ݡȤƤʤ󥳡Ȥޤ"});
        throw new YException(str);
      }
      break;
    };
  }

  if (msg.nRects != 0) {
    wreq_dsp(wid);
  }  
}

// ----------------------------------------------------
//  VNC: ФΥåץܡɼ
// ----------------------------------------------------
// not 'extern'ed as of yet
YString execServerCutText(YLibDraft::YTCPSocket *sock)
{
  rfbServerCutTextMsg msg;
  msg.type = rfbServerCutText; // not needed?
  sock->read((B*)(&msg.pad1),
             sizeof(rfbServerCutTextMsg) - sizeof(char));
  msg.length = ntohl(msg.length);
  unsigned char *buf = new unsigned char[msg.length+1];
  TC *tcs = new TC[msg.length+1];
  buf[msg.length] = '\0';
  sock->read((B*)buf, msg.length);
  sjstotcs(tcs, buf);
  YString str(tcs);
  delete tcs;
  delete buf;
  return str;
}

// ----------------------------------------------------
//  VNC: ФΥåѥ
// ----------------------------------------------------
//  () Ʊ !!
void vncDataReceiver(W arg)
{
  char type;
  VncWindow *win = (VncWindow*)arg;
  
  while (true) {
    try {
      win->sock->read(&type, 1);
      switch (type) {
      case rfbFramebufferUpdate:
        {
          execFramebufferUpdate(win->sock, &(win->disp_sem), win->server, win->windowID, win->mem_gid);
          // rectϰϤʤ٤
          YRect rect(0, 0, win->server->framebufferWidth, win->server->framebufferHeight);
          requestVNCFramebufferUpdate(win->sock, &win->sem, rect, false);
        }
        break;
      case rfbBell:
        sig_buz(0);
        break;
      case rfbServerCutText:
        win->trayString = execServerCutText(win->sock);
        break;
      case rfbSetColourMapEntries:
      default:
        {
          YString str((TC[]){L"ݡȤƤʤåޤ"});
          throw new YException(str);
        }
        break;
      };
    }
    catch (YException *e){
      errorpanel((const TC*)e->getMessage());
      delete e;
    }
  }
  ext_tsk();
}

// ----------------------------------------------------
//  VNC: ɤѴ
// ----------------------------------------------------
unsigned long convertVNCKey(TC keycode)
{
  switch (keycode){
  case KC_BS:
    return 0xff08U; // BS
  case KC_TAB:
    return 0xff09U; // TAB
  case KC_NL:
  case KC_CR:
    return 0xff0dU; // ENTER
  case KC_CAN:
    return 0xff1bU; // ESC
  case INS_CODE:
    return 0xff63U; // INS
  case KC_DEL:
    return 0xffffU; // DEL
  case KC_HOME:
    return 0xff50U; // HOME
  case KC_END:
    return 0xff57U; // END
  case KC_PGUP:
    return 0xff55U; // PGUP
  case KC_PGDN:
    return 0xff56U; // PGDN
  case KC_CC_L:
    return 0xff51U; // LEFT
  case KC_CC_U:
    return 0xff52U; // UP
  case KC_CC_R:
    return 0xff53U; // RIGHT
  case KC_CC_D:
    return 0xff54U; // DOWN
  case KC_SHT_L:
    return 0xffe1U; // SHIFT(left)
  case KC_SHT_R:
    return 0xffe2U; // SHIFT(right)
  case KC_CMD:
    return 0xffe3U; // CTRL(left)
  case KC_EXP:
    return 0xffe9U; // ALT(right)
  // -- followings are designated by the patch created by UnderDone.
  //    some of the mappings are not sure.    
  case KC_JPN0:
    return 0xff25U; // to hiragana
    //return VK_DBE_ROMAN
  case KC_JPN1:
    return 0xff26U; // to katakana
  case KC_CNV0:
    return 0xff22U; // ̵Ѵ
  case KC_CNV:
    return 0xff23U; // Ѵ
  case KC_HAN:
  case KC_EIJI:
    //return 0xff21U; // 
    return 0xff2aU; // /Ⱦ
  case KC_RCNV:
    return 0xff3e; // 
  //case KC_ENGALT:
  //  return VK_DBE_ALPHANUMERIC;
  // -- following keys are not at all sure ...
  //+  { XK_Zenkaku,          VK_DBE_SBCSCHAR, 0 },
  //+  { XK_Hankaku,          VK_DBE_DBCSCHAR, 0 },
  case KC_PRSCRN:
    return 0xff61U; // PRSCRN???
  };
  // F1 - F12
  if ((unsigned long)keycode >= KC_PF1 &&
      (unsigned long)keycode <= KC_PF12){
    return keycode - KC_PF1 + 0xffbeU;
  }
  // usual characters
  unsigned char tmp[3];
  sprintf((char*)tmp, "%C", keycode);
  if (tmp[0] >= 0x20U && tmp[0] <= 0x7fU){
    return tmp[0];
  }
  return 0xffffffU; // void symbol
}

// ----------------------------------------------------
//  VNC: ٥Ȥ
// ----------------------------------------------------
void execVNCKeyEvent(YLibDraft::YTCPSocket *sock, UH key, bool down, UW status)
{
  if (keyTable == 0){
    int size = get_ktb(NULL);
    keyTable = new char[size+1];
    get_ktb((KeyTab*)keyTable);
  }
  // Ѹ졦shiftʳΥ᥿offξ֤ξΥɤ롣
  int sel = 1 + (((status & ES_LSHFT) != 0) ? 4 : 0)
    + (((status & ES_RSHFT) != 0) ? 8 : 0);
  TC keycode = ((KeyTab*)keyTable)->
    kct[((KeyTab*)keyTable)->keymax * ((KeyTab*)keyTable)->kctsel[sel] + key];

  rfbKeyEventMsg msg;
  msg.type = rfbKeyEvent;
  msg.down = (down ? 1 : 0);
  msg.key = htonl(convertVNCKey(keycode));
  sock->write((B*)&msg, sizeof(rfbKeyEventMsg));
}

// ----------------------------------------------------
//  VNC: ü쥭٥Ȥ
// ----------------------------------------------------
void execVNCStatusKeyEvent(YLibDraft::YTCPSocket *sock, UW status, UW oldstatus){
  rfbKeyEventMsg msg;
  msg.type = rfbKeyEvent;
  if ((status & ES_EXT) != (oldstatus & ES_EXT)){
    msg.key = htonl(convertVNCKey(KC_EXP));
    msg.down = (((status & ES_EXT) != 0) ? 1 : 0);
    sock->write((B*)&msg, sizeof(rfbKeyEventMsg));
  }
  if ((status & ES_CMD) != (oldstatus & ES_CMD)){
    msg.key = htonl(convertVNCKey(KC_CMD));
    msg.down = (((status & ES_CMD) != 0) ? 1 : 0);
    sock->write((B*)&msg, sizeof(rfbKeyEventMsg));
  }
  if ((status & ES_RSHFT) != (oldstatus & ES_RSHFT)){
    msg.key = htonl(convertVNCKey(KC_SHT_R));
    msg.down = (((status & ES_RSHFT) != 0) ? 1 : 0);
    sock->write((B*)&msg, sizeof(rfbKeyEventMsg));
  }
  if ((status & ES_LSHFT) != (oldstatus & ES_LSHFT)){
    msg.key = htonl(convertVNCKey(KC_SHT_L));
    msg.down = (((status & ES_LSHFT) != 0) ? 1 : 0);
    sock->write((B*)&msg, sizeof(rfbKeyEventMsg));
  }
}

// ----------------------------------------------------
//  VNC: 饤ȤΥåץܡ
// ----------------------------------------------------
// not 'extern'ed as of yet
void execClientCutText(YLibDraft::YTCPSocket *sock, const std::string& str)
{
  rfbClientCutTextMsg msg;
  msg.type = rfbClientCutText;
  msg.length = htonl(str.length()+1);
  sock->write((B*)&msg, sizeof(rfbClientCutTextMsg));
  sock->write((B*)(str.c_str()), str.length()+1);
}

